#include "stdio.h"
#include "stdlib.h"
#include "mem.h"

#define NUMSAMPLES 2082884

unsigned char m[44 + 2 * NUMSAMPLES];	// 44 bytes for RIFF WAVE header
unsigned char n[NUMSAMPLES];
unsigned long int nsize;
unsigned long int fsize;
unsigned long int begoff;
char cmdbuf[128];
unsigned long int cntCmd = 0;
int picNum = 0;
int savedPos[300];
unsigned long int cntSaved = 0;

#define PICWIDTH 2000		// about 27.7 microseconds - that enough to display one packet
#define PICHEIGHT 256

#define CUTOFF_VALUE 25

unsigned char pic[PICHEIGHT][PICWIDTH];

void
CleanPic (void)
{
  memset (pic, '0', sizeof (pic));
}

void
SetPoint (int x, int y)
{
  if (x >= 0 && x <= (PICWIDTH - 1) && y >= 0 && y <= (PICHEIGHT - 1))
    {
      pic[(PICHEIGHT - 1) - y][x] = '1';
    }
}

void
SaveNamedPic (char *fname)
{
  FILE *f;
  int i, j;

  fprintf (stderr, "saving %s\n", fname);
  f = fopen (fname, "wt");
  if (!f)
    return;
  fprintf (f, "P1\n#\n%d %d\n", PICWIDTH, PICHEIGHT);
  for (i = 0; i < PICHEIGHT; i++)
    fwrite (&pic[i][0], 1, PICWIDTH, f);
  fclose (f);
}

void
SavePic (int num)
{
  FILE *f;
  int i, j;
  char fname[32];
  sprintf (fname, "packet%.3d.pbm", num);
  SaveNamedPic (fname);
}

void
CreatePacketPic (int mpos, int length)
{
  int i, j;

  fprintf (stdout, "pic %.4d width = %.5d\n", picNum, length);
  fflush (stdout);

  CleanPic ();

  for (i = 0; i < length; i++)
    {
      SetPoint (i, n[mpos + i] ? 128 : 0);
    }

/*
for( i=(100-82); i < length; i+=128 )
	{
	for( j=0; j < 255; j++ )
		SetPoint( i, j );
	}
*/

  SavePic (picNum);
  picNum++;
}

void
DecodeManchester (int mpos)	// int length )
{
  int i, j, k;
  char bits[14 + 1];
  char *q;

  if (!cntCmd)
    memset (cmdbuf, 0, sizeof (cmdbuf));

  // printf ("Decoded: ");
  for (k = 0; k < 14; k++)
    {
      int sumA, sumB;
      sumA = 0;
      for (i = 0; i < 64; i++)
	{
	  if (n[mpos + 18 + k * 128 + i])
	    sumA++;
	}
      sumB = 0;
      for (i = 0; i < 64; i++)
	{
	  if (n[mpos + 18 + 64 + k * 128 + i])
	    sumB++;
	}
      bits[k] = (sumA < sumB) ? '1' : '0';
    }
  bits[14] = 0;
  // printf ("S=%.2s T=%.1s A=%.5s CMD=%.6s\n", &bits[0], &bits[2], &bits[3], &bits[8]);
  i = strtoul (bits, &q, 2);
  {
    int t, addr, cmd;
    cmd = i & 63;
    i /= 64;
    addr = i & 31;
    i /= 32;
    t = i & 1;
    printf ("Decoded: T=%d A=%d CMD=%d\n", t, addr, cmd);
    cmdbuf[cntCmd] = cmd;
    cntCmd++;
  };
}

signed short int
GetWord (unsigned char *x)
{
  signed short int r;
  r = x[1];
  r *= 256;
  r |= x[0];
  return r;
}

void
LoadWAV (void)
{
  FILE *f;

  static const char inputFile[] = "mail0004-attachment.bin";

  f = fopen (inputFile, "rb");
  if (!f)
    {
      printf ("can't open %s\n", inputFile);
      exit (1);
    }

  fseek (f, 0, SEEK_END);
  fsize = ftell (f);
  fseek (f, 0, SEEK_SET);

  if (fsize > sizeof (m))
    {
      printf ("file is too big\n");
      exit (1);
    }

  memset (n, 0, sizeof (n));

  fread (m, 1, fsize, f);
  fclose (f);

  begoff = 0;
  if (0 == memcmp (m, "RIFF", 4))
    begoff = 0x2C;
}

void
RemoveStrayDots (void)
{
  unsigned long int i, j;

// pass 1
  for (i = 0; i < nsize; i++)
    {
      int sum;
      // window 50 pixels
      sum = 0;
      for (j = 0; j < 50; j++)
	sum += n[i + j];
      if (sum < 20)
	n[i] = 0;
    }

// pass 2
  for (i = 5; i < nsize; i++)
    {
      int sum;
      sum = 0;
      for (j = 0; j <= 10; j++)
	{
	  sum = sum + n[(i - 5) + j];
	};

      if (n[i] && sum < 4)
	n[i] = 0;
      else if (!n[i] && sum > 6)
	n[i] = 1;
    }

}

void
FindAllPackets (void)
{
  unsigned long int i;
  int state = 0, sum = 0;
  for (i = 0; i < nsize; i++)
    {
      if (n[i] == state)
	{
	  sum++;
	}
      else
	{
	  if (sum > 4000)
	    {
	      printf ("@%.8lX: sum=%d\n", i, sum);
	      savedPos[cntSaved] = i;
	      cntSaved++;
	    }
	  state = !state;
	  sum = 1;
	  //markPos = i;
	}
    }

  if (sum > 4000)
    {
      printf ("@%.8lX: sum=%d\n", i, sum);
      savedPos[cntSaved] = i;
      cntSaved++;
    }
}

void
main (void)
{
  FILE *f;
  unsigned long int i, j;
  signed short int t;

  LoadWAV ();

  for (i = 1, j = 0; i < NUMSAMPLES; i++)
    {
      signed long xx, prev;
      xx = GetWord (&m[begoff + i * 2]) + 32768;
      prev = GetWord (&m[begoff + (i - 1) * 2]) + 32768;
      n[j] = abs ((xx - prev) / 256) > CUTOFF_VALUE ? 1 : 0;
      j++;
    }

  nsize = j;

  RemoveStrayDots ();
  FindAllPackets ();

  picNum = 0;
  for (i = 1; i < cntSaved; i++)
    {
      CreatePacketPic (savedPos[i - 1] - 100, savedPos[i] - savedPos[i - 1]);
      if (1 == (i % 3))		// we only need to decode each third packet
	DecodeManchester (savedPos[i - 1] - 100);	//, savedPos[i]-savedPos[i-1] );
    }

  printf ("The commands: ");
  for (i = 0; i < cntCmd; i++)
    printf ("%c", cmdbuf[i] + '0');
  printf ("\n");

  for (i = 0; i < cntCmd; i += 2)
    printf ("%c", cmdbuf[i] * 10 + cmdbuf[i + 1]);

}
