#include "stdio.h"
#include "stdlib.h"
#include "string.h"

enum
{
  OPEN = 243, ROW0 = 244, ROW1 = 245, ROW2 = 246, ROW3 = 247,	// outputs
  CLK = 248, RESET = 249, COL0 = 250, COL1 = 251, COL2 = 252	// inputs
};


typedef struct _gate
{
  int dst, srcA, srcB;
} _gate;

#define NUMGATES 248
#define NUMWIRES 253

_gate gate[NUMGATES];

#define NUMKEYWORDS 11

const char keyword[NUMKEYWORDS][8] = {
  "NAND", "OPEN", "ROW0", "ROW1", "ROW2", "ROW3", "CLK", "RESET", "COL0",
  "COL1", "COL2"
};

void
LoadNetlist (void)
{
  FILE *f;
  unsigned int i, j, gateIdx;
  char *p;
  int state;
  char s[50];

  static const char inputFile[] = "NETLIST.NAND";

  memset (gate, 0, sizeof (gate));

  f = fopen (inputFile, "rt");
  if (!f)
    {
      printf ("Can't open %s\n", inputFile);
      exit (1);
    }

  while (fgets (s, sizeof (s), f))
    {
      if (strstr (s, "#"))
	continue;
      p = s;

      state = 0;
      while (p)
	{
	  j = 0xFF;
	  for (i = 0; i < NUMKEYWORDS; i++)
	    {
	      if (0 == strncmp (p, keyword[i], strlen (keyword[i])))
		{
		  j = i;
		  break;
		}
	    }

	  if (j != 0)		// ignore keyword 0 (NAND)
	    {

	      j = (0xFF == j) ? atoi (p) : j + 242;

	      if ((j >= NUMWIRES) || ((0 == state) && (j >= NUMGATES)))
		{
		  printf ("Error in netlist\n");
		  exit (1);
		}

	      switch (state)
		{
		case 0:
		  gateIdx = j;
		  gate[gateIdx].dst = j;
		  state++;
		  break;
		case 1:
		  gate[gateIdx].srcA = j;
		  state++;
		  break;
		case 2:
		  gate[gateIdx].srcB = j;
		  state++;
		  break;
		}

	      if (state > 2)
		break;

	    };

	  p = strstr (p, " ");
	  if (p)
	    p++;
	};

    }
  fclose (f);
}


char retbuf[2048];

int level = 0;

char *
Calc (int gateNum)
{
  char tmp[2048];
  char tmp2[2048];

  level++;

  if (gateNum >= 0xBB && gateNum < OPEN)	// these gates are part of flip-flops
    {
      // calculating flip-flops using recursion is not a good idea :)
      printf ("%d? No way!\n", gateNum);
      exit (1);
    }

  do
    {

      if (gateNum >= CLK)
	{
	  // do not calc inputs
	  strcpy (retbuf, keyword[gateNum - 242]);
	  break;
	};

      switch (gateNum)
	{
	  // registers
	case 3:
	case 4:
	case 0x14:
	case 0x18:
	case 0x19:
	case 0x20:
	case 0x21:
	case OPEN:
	  sprintf (retbuf, "REG%d", gateNum);
	  continue;		// i.e. break
	case ROW0:
	case ROW1:
	case ROW2:
	case ROW3:
	  if (level > 1)
	    {
	      sprintf (retbuf, "ROW%d", gateNum - ROW0);
	      continue;
	    }
	  break;
	};

      if (gate[gateNum].srcA == gate[gateNum].srcB)
	{
	  sprintf (tmp, "!(%s)", Calc (gate[gateNum].srcA));
	  strcpy (retbuf, tmp);
	}
      else
	{
	  strcpy (tmp, Calc (gate[gateNum].srcA));
	  strcpy (tmp2, Calc (gate[gateNum].srcB));
	  sprintf (retbuf, "!(%s & %s)", tmp, tmp2);
	};

    }
  while (0);

  level--;

  return retbuf;
}



void
main (int argc, char **argv)
{
  int i;
  if (argc < 2)
    {
      printf ("\nusage: gatecalc GATE# (e.g.  gatecalc 25  or  gatecalc OPEN)\n\n");
      return;
    }

  LoadNetlist ();

  for (i = 1; i < (sizeof (keyword) / sizeof (keyword[0])); i++)
    {
      if (0 == strcmp (&argv[1][0], &keyword[i][0]))
	{
	  printf ("%s = %s\n", keyword[i], Calc (i + 242));
	  return;
	}
    }

  i = atoi (argv[1]);
  printf ("GATE%d = %s\n", i, Calc (i));
}
