#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#define UINT unsigned int

typedef enum CharClass { UPPER = 0, LOWER = 1, DIGIT = 2, MAGIC = 3 } CharClass;

enum { MARK = -1 };

const UINT numCharOfClass[4] = {11, 17, 5, 2 };
#define numUpper numCharOfClass[ UPPER ]
#define numLower numCharOfClass[ LOWER ]
#define numDigits numCharOfClass[ DIGIT ]
#define numMagic numCharOfClass[ MAGIC ]

enum { pwdLen = 35 };

const UINT bitsSetInAllChars[8] = {16,17,19,12,18,24,28,0};
//arrays are indexed from 1 to simplify code
UINT isUniq[1+pwdLen];
char pwd[1+pwdLen+1];
char mask[1+pwdLen][9];
char valid[1+pwdLen][128]; // valid variants for each char
UINT nVar[1+pwdLen];
UINT minBits[1+pwdLen];
UINT maxBits[1+pwdLen];
double total;
UINT figCnt=1;

const char power2[8] = { 1, 2, 4, 8, 16, 32, 64, 128 };

void SetValidChars( UINT posN, char * z ) { strcpy( valid[posN], z ); nVar[posN] = strlen( valid[posN] ); }
UINT BitsSet( char c ) { UINT i, j; for( i=0, j=0; i<8; i++ ) j += !!(c&power2[i]); return j; }
void CharToBits( char c, char * p ) { int i; p[8]=0; for( i=0; i<8; i++ ) p[i] = '0' + !!( c&power2[7-i] ); }

#define IsUpper(x) IsCharOfClass(x,UPPER)
#define IsLower(x) IsCharOfClass(x,LOWER)
#define IsDigit(x) IsCharOfClass(x,DIGIT)

UINT IsCharOfClass( char c, CharClass k )
{
switch(k)
	{
	case UPPER: return (c >= 'A' && c <= 'Z');
	case LOWER: return (c >= 'a' && c <= 'z');
	case DIGIT: return (c >= '0' && c <= '9');
	case MAGIC: return !(IsUpper(c)|IsLower(c)|IsDigit(c));
	}
return 0;
}

UINT PosHasClass( UINT posN, CharClass k )
{ 
UINT i, j; 
if( !nVar[posN] )
	{
	printf( "Error: no valid characters left for position %d\n", posN );
	exit(1);
	}
for( i=0, j=0; i < nVar[posN]; i++ ) j += IsCharOfClass( valid[posN][i], k ); return ( j == nVar[posN] );
}

UINT CountPosOfClass( CharClass k )  { UINT i, j; for( i=1, j=0; i<=pwdLen; i++ ) j += PosHasClass(i,k); return j; }

UINT CheckMask( char c, UINT posN )
{
int i;
char b[9];
CharToBits( c, b );

if( 0 == memcmp( &mask[posN][0], &mask[posN][4], 4 )) // nibbles in the mask are equal
	if( 0 != memcmp( &b[0], &b[4], 4 )) // but in this char, they aren't
		return 0;

for( i=0; i<8; i++ )
	switch( mask[posN][i] ) { case '0': case '1': if( mask[posN][i] != b[i] ) return 0; }

return 1; // ok
}

void AddValidChar( UINT posN, char c )
{
char x[2];
x[0] = c; x[1] = 0;
if( !CheckMask( c, posN )) return;
if( !strstr( valid[posN], x )) { strcat( valid[posN], x ); nVar[posN] = strlen( valid[posN] ); }
}

void RemoveMarked( UINT posN )
{
UINT i, j;
char x[128];
for( i=0, j=0; i < nVar[posN]; i++ ) if( MARK != valid[posN][i] ) x[j++] = valid[posN][i];
x[j] = 0; strcpy( valid[posN], x ); nVar[posN] = j;
}

void ExcludeCharFromValid( UINT posN, char c )
{
char *p;
p = strchr( valid[posN], c ); if(p) { *p = MARK; RemoveMarked( posN ); }
}

void ListAll( void )
{
UINT i, j;
printf( "\n\nFigure #%d.\n", figCnt ); figCnt++;
for( i=1; i <= pwdLen; i++ )
	{
	printf( "char #%.2d - %s - ", i, mask[i] );
	if( minBits[i] == maxBits[i] )
		printf( "%d bits set     ", minBits[i] );
	else
		printf( "%d to %d bits set", minBits[i], maxBits[i] );
	printf( " - %2d variants: [%s]\n", nVar[i], valid[i] ); 
	}

printf( "Uppercase: %d from %d found\n", CountPosOfClass(UPPER), numUpper);
printf( "Lowercase: %d from %d found\n", CountPosOfClass(LOWER), numLower);
printf( "Digits: %d from %d found\n", CountPosOfClass(DIGIT), numDigits);
printf( "Magics: %d from %d found\n", CountPosOfClass(MAGIC), numMagic);

total = 1.0; for( i=1; i <= pwdLen; i++ ) total *= (double)nVar[i];
printf("Total: %5.5f passwords to check\n", total );
}

void ReplaceNibbles( char * pattern, char * repl )
{
UINT i;
char *p;
for( i=1;i <= pwdLen;i++)
	{
	p=&mask[i][0]; if( 0 == memcmp( p, pattern, 4 )) memcpy( p, repl, 4 );
	p+=4; if( 0 == memcmp( p, pattern, 4 )) memcpy( p, repl, 4 );
	}
} 

void UpdateMask( UINT posN )
{
UINT i, j, k;
char unimask[9];
k=nVar[posN];
if(!k)return;

// calc combined mask for the position (from all valid chars)
CharToBits( valid[posN][0], unimask );
for( i=1; i < k; i++ )
	{
	char curmask[9];
	CharToBits( valid[posN][i], curmask );

	for( j=0; j < 8; j++ )
		{
		if( unimask[j] == '?' )
			continue;
		if( unimask[j] != curmask[j] )
			unimask[j] = '?';
		}
	}

// now replace the old mask with the new
for( i=0; i < 8; i++ ) if( '?' == mask[posN][i] ) mask[posN][i] = unimask[i];
}

void UpdateAll( void )
{
UINT i, j, k, N, err=0;

for( N=0; N < 2; N++ ) {

// check again number of bits
for( i=1; i <= pwdLen; i++ )
	{
	for( j=0; j < nVar[i]; j++ )
		{
		k=BitsSet( valid[i][j] );
		if( !(k >= minBits[i] && k <= maxBits[i]) )
			{
			valid[i][j] = MARK;
			}
		else
		if( 0 == CheckMask( valid[i][j], i ))
			{
			valid[i][j] = MARK;
			}
		}
	RemoveMarked( i );
	}

// if char is known exactly and is unique, it cannot be in other positions
for( i=1; i <= pwdLen; i++ )
	if( (1==nVar[i]) && isUniq[i])
		for( j=1; j < pwdLen; j++ ) if( j != i ) ExcludeCharFromValid( j, valid[i][0] );

{
CharClass x;
for( x=UPPER; x <= MAGIC; x++ )
	{
	i = CountPosOfClass(x);
	if( i > numCharOfClass[x] ) { err = 1; };
	if( i == numCharOfClass[x] )
		{
		for( i=1; i <= pwdLen; i++ )
			{
			if( !PosHasClass(i,x))
				{ 
				for( j=0; j < nVar[i]; j++ ) if( IsCharOfClass( valid[i][j], x )) valid[i][j] = MARK;
				RemoveMarked( i );
				}
			}
		}
	}
}

for( i=1; i <= pwdLen; i++ ) UpdateMask(i);
for( i=1; i <= pwdLen; i++ ) if( 1 == nVar[i] ) minBits[i] = maxBits[i] = BitsSet( valid[i][0] ); 
};

ListAll();
if(err)
	{
	printf( "Error: too many chars of certain class(es)\n" );
	exit(1);
	}
}



UINT ValidatePassword( const char * const p )
{
UINT i, j, bitK[8], kU, kL, kD, kO;

kU=kL=kD=kO=0;
memset( bitK,0,sizeof(bitK));
for( i=1; i <= pwdLen; i++ )
	{
	char c=p[i]; 
	if( IsUpper(c)) kU++; 
	else if( IsLower(c)) kL++;
	else if( IsDigit(c)) kD++;
	else kO++;
	for( j=0; j < 8; j++ )
		bitK[j] += !!(c&power2[j]);
	}
if( !(kU == numUpper && kL == numLower && kD == numDigits && kO == numMagic) ) return 0; // check number of letters/digits/etc.
for( i=0; i < 8; i++ ) if( bitK[i] != bitsSetInAllChars[i] ) return 0; // check number of bit set in all chars

// check equality
if( ( p[9]  != p[19] )||( p[13] != p[25] )||( p[29] != p[32] )||( p[8]  != p[28] )||
    ( p[3]  != p[15] )||( p[3]  != p[27] )||( p[16] != p[30] )
) return 0;

// check unique chars
{
char k[256];
memset(k,0,sizeof(k));
for( i=1;i <= pwdLen;i++ )  k[ p[i] ]++;
for( i=1;i <= pwdLen;i++ ) if( isUniq[i] && (k[ p[i] ] != 1)) return 0;
}

// validate SHA256

fprintf( stderr, "." );
{
FILE *f;
char hash[1+64+1];

f=fopen( "tmp", "wb" );
if(!f)return 0;
fwrite(&p[1], 1, pwdLen, f );
fclose(f);
system( "sha256sum tmp > hash" );

f=fopen( "hash", "rb" );
if(!f)return 0;
fread( &hash[1], 1, 64, f );
fclose(f);

if( p[1]    != hash[1]  ) return 0;
if( hash[2] != hash[63] ) return 0;
if( hash[3] != hash[64] ) return 0;
}

printf( "\nVALID PASSWORD: %s\n", &p[1]);
return 1;
}

char currpwd[1+pwdLen+1];
int cnt=0;
int cntValidPwds=0;

void Gen( int pos )
{
UINT i, j, k;
if( pos > pwdLen )
	{
	currpwd[pos]=0;
	if( ValidatePassword(&currpwd[0]))
		{cntValidPwds++; printf( "%.3d" "%c" ": %s\n", (int)(100.0*(double)cnt/total), 0x25, &currpwd[1] );}
	cnt++;
	return;
	}

for( i=0; i < nVar[pos]; i++ )
	{
	currpwd[pos] = valid[pos][i];
	Gen(pos+1);
	}
}


void H(int i, char * p) { memcpy( &mask[i][0], p, 4 ); }
void L(int i, char * p) { memcpy( &mask[i][4], p, 4 ); }

void Init( void )
{
UINT i, j, k;
char *p;

memset(pwd,0,sizeof(pwd));
memset(valid,0,sizeof(valid));

// mark unique chars
{
const int uniqChars[]={ 1, 2, 4, 5, 6, 7, 10, 11, 12, 14, 17, 18, 20, 21, 22, 23, 24, 26, 31, 33, 34, 35 };
for( i=0; i < (sizeof(uniqChars)/sizeof(uniqChars[0])); i++ )
	isUniq[ uniqChars[i] ]=1;
};

// initialize less/more statements 
for( i=1; i <= pwdLen; i++ )
	{
	switch( "?MMLMMMMM.MMLMLLML!.LLMMLMMLM.MM.!LM"[i] )
		{
		case '.': minBits[i] = maxBits[i] = 2; break;
		case 'L': minBits[i] = maxBits[i] = 3; break;
		case 'M': minBits[i] = 4; maxBits[i] = 5; break;
		case '!': minBits[i] = maxBits[i] = 6; break;
		}
	}	

// set masks
for( i=1; i <= pwdLen; i++ ) strcpy( mask[i], "0???????" );

// apply nibble/char equality rules
p="0aaa"; H(6,p); L(23,p); H(10,p); L(6,p); H(20,p); H(26,p);
p="0bbb"; H(8,p); L(11,p); L(17,p); L(21,p); L(29,p); L(32,p); H(24,p); H(28,p); H(29,p); H(31,p); H(32,p); H(34,p);
p="0ccc"; H(9,p); L(13,p); L(16,p); L(25,p); L(30,p); H(13,p); H(17,p); H(19,p); H(25,p);
p="0ddd"; H(11,p); L(10,p); L(18,p); H(14,p); H(16,p); H(18,p); H(22,p); H(23,p); H(30,p);
p="eeee"; L(9,p); L(19,p); // no high nibbles, so high bit is unknown
p="ffff"; L(8,p); L(28,p); // the same
p="0ggg"; H(3,p); H(15,p); H(27,p);
p="hhhh"; L(3,p); L(15,p); L(27,p); // the same

SetValidChars( 1, "0123456789ABCDEFabcdef" ); // char #1 is hex

for( i=2;i <= pwdLen; i++ )
	{
	for(j=32; j < 128; j++ ) // only printable ASCII chars
		{
		k=BitsSet(j);
		if( k >= minBits[i] && k <= maxBits[i] )
			{
			if( IsCharOfClass(j,MAGIC))
				{
				if (j < 0x40)
					AddValidChar(i,j);
				}
			else
				AddValidChar(i,j);
			}
		}
	}
}


#define Say(x) printf( "\n" x "\n" )

void main()
{
UINT i, j, k;

Say("The beginning.");
Init();
ListAll();

Say("Set 0ddd to 0111 (i.e. 7)");
ReplaceNibbles( "0ddd", "0111" ); UpdateAll();

Say("Set 0bbb to 0100 (i.e. 4)");
ReplaceNibbles( "0bbb", "0100" ); UpdateAll();

Say("Set 0ccc to 0cc1 (i.e. 3 or 5)");
Say("Set eeee to 0000" );
ReplaceNibbles( "0ccc", "0cc1" ); ReplaceNibbles( "eeee", "0000" ); UpdateAll();

Say("0aaa is either 3, 5 or 6");
for( i=1; i <= pwdLen; i++ )
	{
	if( 0 == memcmp( mask[i], "0aaa", 4 ))
		{
		for( j=0; j < nVar[i]; j++ )
			{
			k=valid[i][j] / 16;
			switch(k){ case 3: case 5: case 6: break; default: valid[i][j] = MARK; };
			}
		RemoveMarked( i );
		}

	if( 0 == memcmp( &mask[i][4], "0aaa", 4 ))
		{
		for( j=0; j < nVar[i]; j++ )
			{
			k=valid[i][j]&15;
			switch(k){ case 3: case 5: case 6: break; default: valid[i][j] = MARK; };
			}
		RemoveMarked( i );
		}
	}
UpdateAll();

// suppose 0aaa = 3, 0cc1 = 3 -- no valid chars left for #13
// ReplaceNibbles( "0cc1", "0011" ); ReplaceNibbles( "0aaa", "0011" ); UpdateAll();

// suppose 0aaa = 3, 0cc1 = 5 -- too many uppercase chars (12)
// ReplaceNibbles( "0cc1", "0101" ); ReplaceNibbles( "0aaa", "0011" ); UpdateAll();

// suppose 0aaa = 5, 0cc1 = 5 -- no valid chars left for #13
// ReplaceNibbles( "0aaa", "0101" ); ReplaceNibbles( "0cc1", "0101" ); UpdateAll();

// suppose 0aaa = 6, 0cc1 = 5 -- too many uppercase chars (12)
// ReplaceNibbles( "0aaa", "0110" ); ReplaceNibbles( "0cc1", "0101" ); UpdateAll();

// so 0cc1 = 3, 0aaa = 5 or 6 (i.e. 01aa)
Say("Set 0cc1 to 3 and 0aaa to 01aa");
ReplaceNibbles( "0cc1", "0011" ); ReplaceNibbles( "0aaa", "01aa" ); UpdateAll();

// now all digits have been found!
Say("Set the beginning to 'flag<' and last char to '>'");
Say("Also set 0aaa, hhhh and 0ggg");

SetValidChars( 1,  "f" );
SetValidChars( 2,  "l" );
SetValidChars( 3,  "a" ); SetValidChars( 15, "a" ); SetValidChars( 27, "a" );
SetValidChars( 4,  "g" );
SetValidChars( 5,  "<" );
SetValidChars( 35, ">" );

ReplaceNibbles( "01aa", "0101" ); ReplaceNibbles( "hhhh", "0001" );
ReplaceNibbles( "0ggg", "0110" ); UpdateAll();

// now bruteforce password
Say("Bruteforcing password");
Gen(1);
printf( "\nFound %d valid password(s)\n", cntValidPwds );
Say("The end.");
}
